# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .context import context, const_solution_vector
from . import ad
from .solver import _matrixsolve, spsolve_scipy, SolverObject
import numpy as np

__all__ = ['sensitivity_analysis', 'add_sensitivity']


class sensitivity_analysis(SolverObject):
    def __init__(self, sparams, function='J', spsolve=spsolve_scipy):
        self.sensitivity = {}
        self.function = function
        self.sparams = sparams
        self.spsolve = spsolve
        self._g = {}
        self._dg = {}

    def goal(self, solution, time, x, xt, params, full_output):
        return full_output[self.function]

    def process(self, solution):
        n = len(solution.x)
        nparams = self.sparams.values(solution.params)

        def gets(obj):
            if isinstance(obj, const_solution_vector):
                return np.asarray(0.)
            else:
                return self.sensitivity[id(obj)]
        if solution.old:
            nold = sum(o.x * w for o,
                       w in zip(solution.old,
                                solution.weights[1:]))
            dxold = sum(gets(o) *
                        w for o, w in zip(solution.old, solution.weights[1:]))
        else:
            nold = np.zeros_like(solution.model.X)
            dxold = np.asarray(0.)
        if not dxold.shape:
            dxold = dxold * np.ones((n, len(nparams)))
        nvec = np.concatenate([solution.x, nold, nparams])
        vec = ad.forward.seed_sparse_gradient(nvec)
        x = vec[:n]
        xold = vec[n:2 * n]
        xt = solution.weights[0] * x + xold
        p = vec[2 * n:]
        params = self.sparams.asdict(solution.params, p)
        full_output = {}
        F = solution.model.residuals(
            solution.time, x, xt, params, full_output=full_output, solver=self)
        G = self.goal(solution, solution.time, x, xt, params, full_output)

        def split(g):
            g = g.gradient.tocsr()
            return g[:, :n], g[:, n:2 * n], g[:, 2 * n:]
        dF_dx, dF_dxold, dF_dp = split(F)
        dG_dx, dG_dxold, dG_dp = split(G)
        b = -dF_dp - dF_dxold.dot(dxold)
        scaling = solution.model.scaling(solution.params)
        dx_dp = _matrixsolve(self.spsolve, dF_dx, b, scaling)
        self.sensitivity[id(solution)] = dx_dp
        d = dG_dx.dot(dx_dp) + dG_dxold.dot(dxold) + dG_dp
        self._g[id(solution)] = G.value
        self._dg[id(solution)] = d

    def add_hook(self, context):
        return self.process(context.solution)

    def g(self, c):
        return [self._g[id(u)] for u in c.trajectory]

    def dg(self, c):
        return [self._dg[id(u)] for u in c.trajectory]


def add_sensitivity(context, *args, **kwargs):
    sa = sensitivity_analysis(*args, **kwargs)
    context.add_hooks.append(sa)
    return sa



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import scipy.constants
from oedes.models import solver
from .base import EquationWithMesh
from .poisson import Poisson

__all__ = ['ChargedSpecies']


class ChargedSpecies(EquationWithMesh):
    """
    Abstract class for any distribution of charge
    """

    def __init__(self, mesh=None, name=None, z=1, poisson=None, thermal=None):
        if poisson is not None and mesh is None:
            mesh = poisson.mesh
        super(ChargedSpecies, self).__init__(mesh, name)
        self.z = z
        self.poisson = poisson
        self.thermal = thermal

    def newDiscreteEq(self, builder):
        obj = super(ChargedSpecies, self).newDiscreteEq(builder)
        obj.ze = scipy.constants.elementary_charge * self.z
        obj.z = self.z
        obj.mesh = builder.getMesh(self.mesh)
        return obj

    def buildDiscreteEq(self, builder, obj):
        super(ChargedSpecies, self).buildDiscreteEq(builder, obj)
        assert isinstance(self.poisson, Poisson)

    def initDiscreteEq(self, builder, obj):
        super(ChargedSpecies, self).initDiscreteEq(builder, obj)
        obj.poisson = builder.get(self.poisson)
        obj.poisson.species.append(obj)
        obj.poisson.evaluate.depends(obj.load)
        obj.poisson.allspecies.depends(obj.evaluate)
        obj.load.depends(obj.poisson.load)
        obj.alldone.depends(obj.poisson.load)  # keep vars
        obj.alldone.depends(obj.poisson.alldone)
        if self.thermal is not None:
            obj.thermal = builder.get(self.thermal)
            obj.load.depends(obj.thermal.load)
            obj.alldone.depends(obj.thermal.load)  # keep vars

    def load(self, ctx, eq):
        super(ChargedSpecies, self).load(ctx, eq)
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            return
        self.load_concentration(ctx, eq)

    def load_concentration(self, ctx, eq):
        raise NotImplementedError()

    def output_concentration(self, ctx, eq):
        variables = ctx.varsOf(eq)
        c = variables['c']
        ctx.outputCell([eq, 'c'], c, unit=ctx.units.concentration)
        ctx.outputCell([eq, 'charge'], c*eq.ze, unit=ctx.units.charge_density)
        if 'ct' in variables:
            ctx.output([eq, 'ct'], variables['ct'],
                       unit=ctx.units.dconcentration_dt)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import ad


def B(x):
    """
    Bernoulli function B(x)=x/(exp(x)-1), implemented for double precision
    calculation
    """

    # For small x, Taylor expansion should be used instead of direct
    # formula which will is not even defined for x=0.
    BP = 0.0031769106669272879

    # Taylor series
    x2 = x**2.
    taylor = 1. - x / 2. + x2 / 12. - x2**2 / 720.

    # Direct calculation
    # 1e-20 is added to avoid NaN for small x. It only modifies
    # the result if abs(x)<BP. In such a case, the direct formula
    # is not used anyway.
    # In order to avoid overflow errors, maximum x is limited to 700
    # to ensure that exp(xn) is always representable in double precision.
    # Anyway xn/(exp(xn)-1.) is almost 0. for large x
    xn = ad.where(x < 700., x, 700.)
    direct = xn / ((ad.exp(xn) - 1.) + 1e-20)

    # Breakpoint BP is chosen so that Taylor series and direct
    # calculation coincide in double precision.
    return ad.where(abs(x) < BP, taylor, direct)


def Aux1(x):
    """Auxiliary function Aux1(x) = x/sinh(x)
    Properties:
    Aux1(x)=Aux1(-x)
    Aux1(x/2)-x*Aux2(x/2)=B(x)
    """
    return Aux2(-x) * B(x) * 2.


def Aux2(x):
    """
    Auxiliary function Aux2(x) = 1./(exp(x)+1)
    Properties: Aux2(x)+Aux2(-x)=1
    """
    xn = ad.where(x < 700., x, 700.)
    return 1. / (ad.exp(xn) + 1.)


def ScharfetterGummelFlux(dr, cfrom, cto, v, D, full_output=False):
    """
    Calculate Scharfetter-Gummel approximation of flux between points
    separated by distance dr, with species concentrations cfrom, cto,
    drift velocity v and diffusion coefficient D.
    Returns: flux or (flux,c,dc/dx) in center, if full_output
    """
    u = D / dr
    t = v / u
    Bt = B(t)
    flux = (cfrom * B(-t) - cto * Bt) * u
    if not full_output:
        return flux
    t2 = 0.5 * t
    aux2 = Aux2(t2)
    aux1 = Bt + t * aux2
    c = cfrom * (1. - aux2) + cto * aux2
    dc = aux1 * (cto - cfrom) / dr
    return (flux, c, dc)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import scipy.constants
from oedes import ad
import numpy as np


def LangevinRecombination(mu_n, mu_p, n, p, epsilon, npi):
    "Langevin recombination rate"
    return (n * p - npi) * scipy.constants.elementary_charge / \
        epsilon * (mu_n + mu_p)


def EmtageODwyerBarrierLowering(F, epsilon):
    """
    Emtage-O'Dwyer barrier lowering formula
    F must be positive
    """
    k = scipy.constants.elementary_charge / (4 * np.pi * epsilon)
    return np.sqrt(F * k)


def ThermalVoltage(T):
    return scipy.constants.Boltzmann * T / scipy.constants.elementary_charge


def spacing_to_N0(a):
    return a**-3.


def N0_to_spacing(N0):
    return N0**-(1. / 3.)


def MottGurney(epsilon_r, mu, V, L):
    return 9. / 8. * scipy.constants.epsilon_0 * epsilon_r * mu * V**2 / L**3


ManyRakavy_t1 = 2. * (1. - np.exp(-0.5))


def ManyRakavy(mu, V, L):
    return ManyRakavy_t1 * L**2 / (mu * V)


def OnsagerFunction(b):
    """
    Onsager function of real variable J1(2 \sqrt(-2b))/sqrt(-2b), b>=0
    """
    def f(x):
        return scipy.special.i1(2. * x) / x

    def df(args, f):
        x, = args
        u = 2. * x
        yield lambda: 2. * (scipy.special.i0(u) - f) / x
    return ad.custom_function(f, df)(np.sqrt(2. * b + 1e-10))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import scipy.sparse
import scipy.constants
from oedes.ad import isscalar, stack, dot
from collections import OrderedDict


class meshpart(object):
    "Part of mesh"
    cells = None
    idx = None
    fluxsum = None

    def __init__(self, idx=None):
        self.idx = idx
        self.default_electrode = None

    def setUp(self):
        pass

    @property
    def ncells(self):
        return len(self.cells)

    @property
    def x(self):
        return self.cells['center']


class mesh(meshpart):
    "Vertex centered finite volume mesh"
    cells = None
    faces = None
    fluxsum = None

    internal = None
    boundary = None

    group = None

    def getSharedBoundary(self, other, name):
        assert self.group is not None, 'must belong to group'
        return self.group.getSharedBoundary(self, other, name)

    def __init__(self):
        meshpart.__init__(self)

    def facegrad(self, x):
        return (x[self.faces['+']] - x[self.faces['-']]) / self.faces['dr']

    def faceaverage(self, x):
        if isscalar(x):
            return x
        else:
            return (x[self.faces['-']] + x[self.faces['+']]) * 0.5

    def cellaveragev(self, x):
        # unimplemented
        return np.NaN

    def magnitudev(self, x):
        # unimplemented
        return np.NaN

    def dotv(self, x, y):
        # unimplemented
        return np.NaN

    def setUp(self):
        _internal = np.ones(self.ncells, dtype=np.bool)
        _internal[self.boundary.idx] = False
        _internal_idx = np.arange(self.ncells)[_internal]
        self.internal = meshpart(_internal_idx)
        assert (np.unique(self.boundary.idx) == self.boundary.idx).all(
        ), "bad boundary subset : idx must be sorted and unique"
        assert len(self.internal.idx) + len(
            self.boundary.idx) == self.ncells, "bad boundary specification : duplicate idx"
        for obj in [self.internal, self.boundary]:
            obj.fluxsum = self.fluxsum[obj.idx]
            obj.cells = self.cells[obj.idx]
            obj.setUp()

    @property
    def ndim(self):
        c = self.cells['center']
        if len(c.shape) == 1:
            return 1
        else:
            return c.shape[1]


class mesh1d_group(object):

    def getSharedBoundary(self, mesh, other, name):
        assert other.group is self and mesh.group is self
        assert name in mesh.boundaries, 'boundary not known'
        other, = (k for k in mesh.boundaries if k != name)
        return mesh.boundaries[name], mesh.boundaries[other]


class mesh1d(mesh):
    "Unidimensional mesh"

    cellavg = None

    def spacing(self, L, n_desired=200, dx_boundary=1e-10):
        dx_dl = dx_boundary
        dx_desired = L / n_desired

        # Spacing for double layer region
        dldx_ = 10**np.linspace(np.log10(dx_dl), np.log10(dx_desired),
                                int(np.ceil(np.log10(dx_desired / dx_dl) * 10)))
        xdl_ = stack(0., np.cumsum(dldx_))
        n_normal_ = (L - (2 * xdl_[-1])) / dx_desired
        assert n_normal_ > 100
        xnorm_ = np.linspace(xdl_[-1], L - xdl_[-1], int(n_normal_))
        x_ = stack(xdl_, xnorm_[1:-1], L - xdl_[::-1])
        return x_

    def __init__(self, L, epsilon_r=3., boundary_names=[
                 'electrode0', 'electrode1'], **kwargs):
        mesh.__init__(self)

        if np.asarray(L).shape == ():
            x_ = self.spacing(L, **kwargs)
        else:
            x_ = L
            L = x_[-1]

        n = len(x_)

        # create cells
        cells = np.zeros(n, dtype=[('center', '1f8'),
                                   ('volume', 'f8'),
                                   ('epsilon', 'f8')])
        # cells['center']=np.linspace(0,L,n)
        cells['center'] = x_
        cells['epsilon'] = scipy.constants.epsilon_0 * epsilon_r
        _dx = np.diff(cells['center']) / 2
        cells['volume'] = np.hstack((_dx, 0)) + np.hstack((0, _dx))
        assert np.allclose(sum(cells['volume']), L)
        self.cells = cells
        assert np.allclose(sum(cells['volume']), L)

        # create faces
        faces = np.zeros(n - 1, dtype=[('-', 'i4'), ('+', 'i4'),
                                       ('surface', 'f8'),
                                       ('dR', '1f8'),
                                       ('dr', 'f8'),
                                       ('epsilon', 'f8'),
                                       ('center', '1f8')])
        faces['-'] = np.arange(0, n - 1)
        faces['+'] = np.arange(1, n)
        faces['surface'] = 1.
        faces['epsilon'] = 0.5 * \
            (cells['epsilon'][faces['-']] + cells['epsilon'][faces['+']])
        faces['center'] = 0.5 * \
            (cells['center'][faces['-']] + cells['center'][faces['+']])
        faces['dR'] = cells['center'][faces['+']] - cells['center'][faces['-']]
        faces['dr'] = np.sqrt(np.sum(np.atleast_2d(faces['dR']**2), axis=0))
        assert np.allclose(sum(faces['dr']), L)
        self.faces = faces

        # create boundary
        self.boundary = meshpart(idx=np.asarray([0, n - 1]))

        self.electrode0 = np.zeros(
            1, dtype=[('bidx', 'i4'), ('normal', '1f8'), ('surface', 'f8')])
        self.electrode1 = np.zeros(
            1, dtype=[('bidx', 'i4'), ('normal', '1f8'), ('surface', 'f8')])
        self.electrode0['bidx'] = 0
        self.electrode1['bidx'] = 1
        self.electrode0['normal'] = 1.
        self.electrode1['normal'] = -1.
        self.electrode0['surface'] = 1.
        self.electrode1['surface'] = 1.
        lname, rname = boundary_names
        self.boundaries = {lname: self.electrode0, rname: self.electrode1}
        self.group = mesh1d_group()

        # create flux summing matrix
        def _fluxsum(cells, fluxes):
            i = np.arange(len(fluxes))
            shape = (cells.shape[0], fluxes.shape[0])

            def m(sign):
                return scipy.sparse.csr_matrix(
                    (fluxes['surface'], (fluxes[sign], i)), shape=shape)
            return m('+') - m('-')

        self.fluxsum = _fluxsum(cells, faces)
        self.fluxsum.sort_indices()

        # create flux averaging matrix
        def _cellsum(fluxsum):
            a = fluxsum.tocoo()
            i, j = a.row, a.col
            A = scipy.sparse.csr_matrix(
                (np.ones(len(i)), (i, j)), shape=(len(self.cells), len(self.faces)))
            s = A.sum(axis=1).A1
            sc = scipy.sparse.diags(
                [1. / s], offsets=[0], shape=(len(self.cells), len(self.cells)))
            return (sc.dot(A)).tocsr()
        self.cellavg = _cellsum(self.fluxsum)
        self.default_electrode = boundary_names[-1]
        self.setUp()

    @property
    def length(self):
        return self.cells['center'][-1]

    def hasBoundary(self, name):
        return name in self.boundaries

    def getBoundary(self, name):
        return self.boundaries[name]

    def cellaveragev(self, x):
        if isscalar(x):
            return x
        return dot(self.cellavg, x)

    def magnitudev(self, x):
        return np.abs(x)

    def dotv(self, x, y):
        return x * y


def emptyBoundary():
    return np.zeros(
        0, dtype=[('bidx', 'i4'), ('normal', '1f8'), ('surface', 'f8')])


class multilayer1d_group:

    def getSharedBoundary(self, mesh, other, name):
        assert other.group is self and mesh.group is self
        assert name in mesh.boundaries, 'boundary not known'
        if other.layer_i == mesh.layer_i + \
                1 and mesh.boundaries[name] is mesh.electrode1:
            return (mesh.electrode1, other.electrode0)
        if other.layer_i == mesh.layer_i - \
                1 and mesh.boundaries[name] is mesh.electrode0:
            return (mesh.electrode0, other.electrode1)
        return None


class multilayer1d:

    def __init__(self, meshes):
        x = 0.
        self.domains = OrderedDict()
        self.group = multilayer1d_group()
        for i, (name, mesh) in enumerate(meshes):
            assert mesh.cells['center'][0] == 0, 'not starting at x=0'
            assert mesh.cells[
                'center'][-1] >= np.amax(mesh.cells['center'][-1])
            assert isinstance(
                mesh.group, mesh1d_group), 'can only belong to trivial group'
            mesh.group = self.group
            mesh.layer_i = i
            L = mesh.cells['center'][-1]
            mesh.cells['center'] += x
            mesh.internal.cells['center'] += x
            mesh.boundary.cells['center'] += x
            mesh.faces['center'] += x
            x += L
            self.domains[name] = mesh



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['BandEnergy', 'BandEnergyFromParams', 'WithDOS', 'DOS']

from oedes.utils import SubEquation
from .charged import ChargedSpecies
from oedes.models import solver
from oedes.ad import getitem


class BandEnergy(SubEquation):
    "Abstract class for calculation of band energy"

    def energy(self, parent, ctx, parent_eq):
        raise NotImplementedError()


class BandEnergyFromParams(BandEnergy):
    "Band energy as `.energy`` parameter"

    def energy(self, parent, ctx, parent_eq):
        return ctx.param(parent_eq, 'energy')


class WithDOS(ChargedSpecies):
    """
    Charged species with DOS

    Outputs:
    --------
    Ef : cell variable
        quasi Fermi energy [ eV ]
    phi_f : cell variable
        quasi Fermi potential [V]
    Eband : cell variable
        band energy [ eV ]
    phi_band : cell variable
        band equivalent potential (Eband/-q)

    """

    def __init__(self, *args, **kwargs):
        dos = kwargs.pop('dos', None)
        band = kwargs.pop('band', BandEnergyFromParams())
        super(WithDOS, self).__init__(*args, **kwargs)
        self.dos = dos
        self.band = band

    def load_concentration(self, ctx, eq):
        super(WithDOS, self).load_concentration(ctx, eq)
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            return
        if self.dos is None:  # TODO
            return
        potential = ctx.varsOf(eq.poisson)['potential']
        Ebandv = self.band.energy(self, ctx, eq)
        Eband = -potential + Ebandv
        ctx.varsOf(eq).update(Ebandv=Ebandv, Eband=Eband)
        Ef = self.dos.QuasiFermiLevel(ctx, eq)
        ctx.varsOf(eq).update(Ef=Ef)
        if ctx.wants_output:
            ctx.outputCell([eq, 'Ef'], Ef, unit=ctx.units.eV)
            ctx.outputCell([eq, 'Eband'], Eband, unit=ctx.units.eV)
            ctx.outputCell([eq, 'phi_f'], -Ef, unit=ctx.units.V)
            ctx.outputCell([eq, 'phi_band'], -Eband, unit=ctx.units.V)

    def buildDiscreteEq(self, builder, obj):
        super(WithDOS, self).buildDiscreteEq(builder, obj)
        obj.dos = self.dos
        obj.band = self.band

    def initDiscreteEq(self, builder, obj):
        super(WithDOS, self).initDiscreteEq(builder, obj)
        if self.dos is None:  # TODO
            return
        self.band.subInit(self, builder, obj)
        self.dos.subInit(self, builder, obj)


class DOS(SubEquation):
    def Ef(self, ctx, eq):
        "Return band-referenced Fermi level from concentration"
        raise NotImplementedError()

    def Ef_correction(self, ctx, eq):
        "Return Ef - kT log c"
        raise NotImplementedError()

    def c(self, ctx, eq, Ef, numeric_parameters):
        "Return concentration from band-referenced Fermi level"
        raise NotImplementedError()

    def QuasiFermiLevel(self, ctx, eq):
        assert eq.z in [1, -1]
        Eband = ctx.varsOf(eq)['Eband']
        return Eband - eq.z * self.Ef(ctx, eq)

    def _concentration(self, ctx, eq, idx, imref, ref, **kwargs):
        assert eq.z in [1, -1]
        if idx is not None:
            Eband = getitem(ctx.varsOf(eq)[ref], idx)
        else:
            Eband = ctx.varsOf(eq)[ref]
        return self.c(ctx, eq, -(imref - Eband) / eq.z, **kwargs)

    def concentration(self, ctx, eq, idx, imref, **kwargs):
        return self._concentration(ctx, eq, idx, imref, 'Eband', **kwargs)

    def concentrationv(self, ctx, eq, idx, imref, **kwargs):
        return self._concentration(ctx, eq, idx, imref, 'Ebandv', **kwargs)

    def D_mu(self, parent_eq, ctx, eq):
        "Return D/mu"
        raise NotImplementedError()



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import matplotlib
import matplotlib.pylab as plt
import matplotlib.ticker
from collections import defaultdict
from oedes.fvm import mesh1d
import numpy as np
import itertools
from matplotlib.collections import LineCollection
from matplotlib.ticker import EngFormatter
from oedes.utils import meta_key, Units

__all__ = ['matplotlib', 'plt', 'forcontext', 'subplots', 'Settings']


class mesh1d_mpl(object):
    def __init__(self, fig, ax):
        self.fig = fig
        self.ax = ax
        self.cycler = itertools.cycle(matplotlib.rcParams['axes.prop_cycle'])
        self.yunits = set()
        self.xunits = set()

    def plot(self, items, *args, **kwargs):
        if 'color' not in kwargs:
            kwargs['color'] = next(self.cycler)['color']
        lines = []
        for item in items:
            meta, value = item
            mesh = meta.mesh
            face = meta.face
            if not isinstance(mesh, mesh1d):
                raise ValueError('mesh type not supported')
            if face:
                X = mesh.faces['center']
            else:
                X = mesh.cells['center']
            lines.append(np.vstack((X, value)).transpose())
            self.yunits.add(meta.unit)
        self.xunits.add(Units.meter)
        self.ax.add_collection(LineCollection(lines, *args, **kwargs))
        self.ax.autoscale()

    def settings(self, settings):
        yunits = ' / '.join(r'$\mathrm{{{}}}$'.format(u)
                            for u in sorted(self.yunits))
        xunits = ' / '.join(r'$\mathrm{{{}}}$'.format(u)
                            for u in sorted(self.xunits))

        if 'ylabel' not in settings:
            self.ax.set_ylabel(yunits)
        if 'xlabel' not in settings and 'xunit' not in settings:
            if self.xunits == set([Units.meter]):
                self.ax.set_xlabel('Position')
                self.ax.xaxis.set_major_formatter(EngFormatter(unit=xunits))
            else:
                self.ax.set_xlabel(xunits)
        for name in ['xlabel', 'ylabel', 'xlim', 'ylim', 'xscale', 'yscale']:
            if name in settings:
                getattr(self.ax, 'set_' + name)(settings[name])

        prefixes = {'n': 1e-9, r'\mu': 1e-6, 'u': 1e-6, 'm': 1e-3, None: 1.}

        def aunit(axobj, unit):
            try:
                pv = prefixes[unit]
            except BaseException:
                pv = float(unit)
            axobj.set_major_formatter(matplotlib.ticker.FuncFormatter(
                lambda x, pos: '{0:g}'.format(x / pv)))
            axobj.set_minor_formatter(matplotlib.ticker.NullFormatter())

        if 'xunit' in settings:
            aunit(self.ax.xaxis, settings['xunit'])
        if 'yunit' in settings:
            aunit(self.ax.yaxis, settings['yunit'])


import re


class Settings:
    """
    Plot presets, following naming
    yquantity[_yunit-if-not-basic]_xquantity[_xunit_if_not_basic][_[log_or_lin_y][log_or_lin_x]]
    """

    c_x_nm_loglin = dict(
        yscale='log',
        ylabel=r'$\mathrm{m^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_cm_x_nm_loglin = dict(
        yscale='log',
        yunit=1e6,
        ylabel=r'$\mathrm{cm^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_x_nm = dict(
        ylabel=r'$\mathrm{m^{-3}}$',
        xunit='n',
        xlabel='nm')
    c_cm_x_nm = dict(
        yunit=1e6,
        ylabel=r'$\mathrm{cm^{-3}}$',
        xunit='n',
        xlabel='nm')
    potential_x_nm = dict(ylabel=r'V', xunit='n', xlabel='nm')
    E_x_nm = dict(ylabel=r'$\mathrm{V \ m^{-1}}$', xunit='n', xlabel='nm')
    Ef_x_nm = dict(ylabel=r'eV', xunit='n', xlabel='nm')

    j_V = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel='V')
    j_V_loglin = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='V',
        yscale='log')
    j_V_loglog = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='V',
        yscale='log',
        xscale='log')

    j_mAcm_V = dict(yunit=10, ylabel=r'$\mathrm{mA \ cm^{-2}}$', xlabel='V')

    j_t = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel='s')
    j_t_loglog = dict(
        ylabel=r'$\mathrm{A \ m^{-2}}$',
        xlabel='s',
        xscale='log',
        yscale='log')
    j_t_us = dict(ylabel=r'$\mathrm{A \ m^{-2}}$', xlabel=r'$\mu s$', xunit='u')
    j_mAcm_t_us = dict(
        yunit=10,
        ylabel=r'$\mathrm{mA \ cm^{-2}}$',
        xlabel=r'$\mu s$',
        xunit='u')


class context_plotting:
    def __init__(self, ctx, fig=None, ax=None):
        self.out = ctx.output()
        self.model = ctx.model
        self.ctx = ctx
        self.fig = fig
        if ax is not None:
            self.ax = ax
        else:
            self.fig, self.ax = plt.subplots()
        self._impl = None

        self.default_settings_species = Settings.c_x_nm_loglin
        self.default_settings_potential = Settings.potential_x_nm
        self.default_settings_E = Settings.E_x_nm
        self.default_settings_Ef = Settings.Ef_x_nm
        self.default_settings_selector = dict()

    def getimpl(self, mesh):
        if self._impl is None:
            if mesh is None or isinstance(mesh, mesh1d):
                c = mesh1d_mpl
            elif hasattr(mesh, 'mpl_class'):
                c = mesh.mpl_class
            else:
                raise ValueError(
                    'do not know mpl_class of %s' %
                    mesh.__class__)
            self._impl = c(self.fig, self.ax)
        return self._impl

    def select(self, expr):
        r = re.compile(expr)
        grouped = defaultdict(list)
        with_label = 'label' in r.groupindex
        for k in sorted(self.out.keys()):
            match = r.match(k)
            if match:
                ent = (self.out[meta_key][k], self.out[k])
                if not with_label:
                    yield k, [ent]
                else:
                    grouped[match.group('label')].append(ent)
        for k in sorted(grouped.keys()):
            yield k, grouped[k]

    def _plot(self, parts, settings=None, *args, **kwargs):
        if not parts:
            return
        impl = self.getimpl(parts[0][0].mesh)
        impl.plot(parts, *args, **kwargs)
        if settings is not None:
            impl.settings(settings)

    def apply_settings(self, settings):
        if self._impl is not None:
            self._impl.settings(settings)

    def plot(self, keys, *args, **kwargs):
        plotlist = [(self.out[meta_key][k], self.out[k]) for k in keys]
        return self._plot(plotlist, *args, **kwargs)

    def selector(self, expr, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_selector
        want_label = 'label' not in kwargs
        for legendkey, parts in self.select(expr):
            if want_label:
                kwargs['label'] = legendkey
            self._plot(parts, *args, **kwargs)

    def allspecies(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_species
        self.selector(r'(.*\.)?(?P<label>.*)\.c$', *args, **kwargs)

    def species(self, prefix, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_species
        self.selector(r'(.*\.)?%s\.c$' % re.escape(prefix), *args, **kwargs)

    def potential(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_potential
        self.selector(r'(.*\.)?(?P<label>potential)$', *args, **kwargs)

    def E(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_E
        self.selector(r'(.*\.)?(?P<label>E)$', *args, **kwargs)

    def Ef(self, *args, **kwargs):
        if 'settings' not in kwargs:
            kwargs['settings'] = self.default_settings_Ef
        self.selector(r'(.*\.)?(?P<label>.*)\.Ef$', *args, **kwargs)

    def teval(self, *functions, **kwargs):
        settings = kwargs.pop('settings', None)
        results = self.ctx.teval(*functions)
        want_label = 'label' not in kwargs
        for i in range(1, len(results)):
            if want_label:
                kwargs['label'] = str(functions[i])
            self.ax.plot(results[0], results[i], **kwargs)
        self.ax.set_xlabel(str(functions[0]))
        if settings:
            self.getimpl(None).settings(settings)

    def apply(self, settings):
        self.impl.apply(settings)


def subplots(c, *args, **kwargs):
    fig, u = plt.subplots(*args, **kwargs)

    def wrap(x):
        if isinstance(x, tuple):
            return tuple(wrap(y) for y in x)
        elif isinstance(x, list):
            return list(wrap(y) for y in x)
        else:
            return context_plotting(c, fig, x)
    return wrap(u)


def forcontext(*args, **kwargs):
    return context_plotting(*args, **kwargs)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.models import BaseModel, electronic_device
from oedes.fvm import mesh1d
from oedes import solve
import numpy as np
from numpy.testing import assert_allclose


def make_simple_device(voltage=0., polarity='p',
                       x=np.linspace(0., 100e-9, 10)):
    model = BaseModel()
    mesh = mesh1d(x)
    electronic_device(model, mesh, polarity)
    params = {'T': 300., 'electrode0.voltage': voltage, 'electrode0.workfunction': 0, 'electrode1.voltage': 0, 'electrode1.workfunction': 0,
              'hole.N0': 1e27, 'electron.N0': 1e27, 'hole.mu': 1e-9, 'electron.mu': 1e-9, 'electron.energy': 0, 'hole.energy': 0, 'epsilon_r': 3.}
    model.setUp()
    return model, mesh, params


def check(model, params, v, p, res1, res2, dtype=np.double):
    x = np.array(model.X, dtype=dtype)
    x[model.poisson.idx] = v
    x[model.species[0].idx] = p

    def residuals(x):
        return model.residuals(0., x, 0. * x, params)

    def resnorm(x):
        return np.linalg.norm(residuals(x))
    assert resnorm(x) < res1
    xr = solve(model, x, params, maxiter=3)
    assert resnorm(xr) < res2


def test_0v():
    v = np.array([0., 0.21687333, 0.26445226, 0.28515601, 0.29379431,
                  0.29379431, 0.28515601, 0.26445226, 0.21687333, 0.])
    p = np.array([1.00000000e+27, 2.27345935e+23, 3.60907604e+22,
                  1.62027335e+22, 1.16003929e+22, 1.16003929e+22,
                  1.62027335e+22, 3.60907604e+22, 2.27345935e+23,
                  1.00000000e+27])
    model, _, params = make_simple_device()
    check(model, params, v, p, 2e24, 1e16)
    check(model, params, v, p, 2e24, 1e16, dtype=np.longdouble)


def test_1v():
    v = np.array([1., 1.20799767, 1.20085886, 1.12804101, 1.01286933,
                  0.86470611, 0.68875172, 0.48838939, 0.26602888, 0.])
    p = np.array([1.00000000e+27, 2.88907411e+23, 8.82005642e+22,
                  5.68770876e+22, 4.43044288e+22, 3.73208501e+22,
                  3.27774890e+22, 2.95414125e+22, 5.86423921e+22,
                  1.00000000e+27])
    model, _, params = make_simple_device(1.)
    check(model, params, v, p, 2e24, 1e18)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import ad
import numpy as np
import scipy.special
import scipy.integrate
import scipy.interpolate
import scipy.version
from packaging.version import Version

if Version(scipy.version.version) < Version('0.19.0'):
    class make_interp_spline:
        def __init__(self, x, y, k=3):
            self.obj = scipy.interpolate.UnivariateSpline(x, y, s=0, k=k)

        def __call__(self, x, nu=0):
            if np.asarray(x).dtype == np.longdouble:
                x = np.asarray(x, dtype=np.double)
            return self.obj.__call__(x, nu=nu)
else:
    make_interp_spline = scipy.interpolate.make_interp_spline


class GaussFermiBase_(object):
    """
    Abstract implementation of Gauss-Fermi integral:
    I(a,b) = 1/sqrt(pi) \int_{-\infty}{+\infty} dx exp(-x^2)/(1.+exp(a*x+b))

    For simulations, three functions have to be provided in a consistent way:
    - forward calculation, provided by method I(a,b)
    - inverse calculation solving I(a,b)=I for b, provided by method b(a,I)
    - derivative calculation dI/db, provided by method dIdb(a,b)

    The last one is necessary because sparsegrad support is limited to one order
    of differentiation, therefore a function calculating first derivative has to
    be explicitly called to enable automatic calculation of second-order
    derivative.

    It is quite hard to give practical approximation of the integral which would
    work for all values. Therefore range of supported parameters values is
    also specified:
    - 0<=a<=a_max (for all calculations)
    - I_min<=I<=I_max (for calculation of b solving I(a,b)=I)

    Functions with underscore operate on internal representation of I, v=f(I).
    """

    maxiter = 20
    a_max = None
    I_min = None
    I_max = None
    atol = 0.
    rtol = 0.

    def _nevaluate(self, a, b, need_value=False, need_da=False,
                   need_db=False, need_d2aa=False, need_d2ab=False, need_d2bb=False):
        """
        Evaluate F (if need_value), and its derivatives (for example, dF/da if need_da and return 'da').
        """
        raise NotImplementedError()

    def _halley(self, function, b0, tol, maxiter):
        """
        Solve equation F(a,b)=i for b. function should return F, F', F'' or 0.
        """
        b = b0.copy()
        for itno in range(maxiter):
            f, df, d2f = function(b)
            if np.all(np.abs(f) < tol):
                return b
            delta = 2. * f * df / (2. * df * df - f * d2f)
            if not np.isfinite(delta).all():
                raise RuntimeError('numeric error')
            b -= delta
        raise RuntimeError('iteration limit exceeded')

    def _guess_b(self, a, v):
        raise NotImplementedError()

    def _nsolve_b(self, a, v, b0=None):
        tol = self.rtol * v + self.atol
        if b0 is None:
            b0 = self._guess_b(a, v)

        def function(b):
            e = self._nevaluate(
                a,
                b,
                need_value=True,
                need_db=True,
                need_d2bb=True)
            return e['value'] - v, e['db'], e['d2bb']
        return self._halley(function, b0, tol, self.maxiter)

    def _value(self, a, b):
        v = self._nevaluate(
            ad.nvalue(a), ad.nvalue(b), need_value=True, need_da=isinstance(
                a, ad.forward.value), need_db=isinstance(
                b, ad.forward.value))

        def _I(a, b):
            return v['value']

        def _I_deriv(args, value):
            yield lambda: v['da']
            yield lambda: v['db']
        return ad.custom_function(_I, _I_deriv)(a, b)

    def _dvdb(self, a, b):
        v = self._nevaluate(
            ad.nvalue(a),
            ad.nvalue(b),
            need_value=False,
            need_db=True,
            need_d2ab=isinstance(
                a,
                ad.forward.value),
            need_d2bb=isinstance(
                b,
                ad.forward.value))

        def _dIdb(a, b):
            return v['db']

        def _dIdb_deriv(args, value):
            yield lambda: v['d2ab']
            yield lambda: v['d2bb']
        return ad.custom_function(_dIdb, _dIdb_deriv)(a, b)

    def _b(self, a, v, b0=None):
        b = self._nsolve_b(ad.nvalue(a), ad.nvalue(v), b0=b0)

        def _b(a, v):
            return b

        def _b_deriv(args, value):
            a_, i_ = args
            b = value
            values = self._nevaluate(
                a_, b, need_da=isinstance(
                    a, ad.forward.value), need_db=True)
            dbdI = 1. / values['db']
            yield lambda: -dbdI * values['da']
            yield lambda: dbdI
        return ad.custom_function(_b, _b_deriv)(a, v)

    def __call__(self, a, b):
        return self.I(a, b)


class GaussFermiBase(GaussFermiBase_):
    rtol = 1e-14
    atol = 0.

    def _guess_b(self, a, i):
        # 1e-20 is added to avoid NaN, has no effect on calculation
        return np.where(i < 0.1, -np.log(i) + 0.25 * a **
                        2, np.log(1.0 / (i + 1e-20) - 1.))

    def I(self, a, b):
        return self._value(a, b)

    def dIdb(self, a, b):
        return self._dvdb(a, b)

    def b(self, a, i, b0=None):
        return self._b(a, i, b0=b0)


class GaussFermiLogBase(GaussFermiBase_):
    atol = 1e-13
    rtol = 0.

    def I(self, a, b):
        v = self._value(a, b)
        l_max = np.log(self.I_max)
        v = ad.where(v <= l_max, v, l_max)
        return ad.exp(v)

    def b(self, a, i, b0=None):
        return self._b(a, ad.log(i), b0=b0)

    def dIdb(self, a, b):
        v = self._nevaluate(
            ad.nvalue(a), ad.nvalue(b), need_value=True, need_da=isinstance(
                a, ad.forward.value), need_db=True, need_d2ab=isinstance(
                a, ad.forward.value), need_d2bb=isinstance(
                b, ad.forward.value))
        i = ad.exp(v['value'])

        def _dIdb(a, _):
            return i * v['db']

        def _dIdb_deriv(args, value):
            yield lambda: i * (v['d2ab'] + v['da'] * v['db'])
            yield lambda: i * (v['d2bb'] + v['db'] * v['db'])
        return ad.custom_function(_dIdb, _dIdb_deriv)(a, b)


class GaussFermiIntegralH(GaussFermiBase):
    """GaussFermiIntegral implemented using Gauss-Hermite quadrature"""

    def __init__(self, n, a_max):
        super(GaussFermiIntegralH, self).__init__()
        self.maxiter = 20
        self.a_max = a_max
        self.I_min = 1e-30
        self.I_max = 0.9999
        _x, _w = scipy.special.h_roots(n)
        self.X_W = list(zip(_x, _w / np.sqrt(np.pi)))

    def _nevaluate(self, a, b, need_value=False, need_da=False,
                   need_db=False, need_d2aa=False, need_d2ab=False, need_d2bb=False):
        I, da, db, d2aa, d2ab, d2bb = None, None, None, None, None, None
        if need_value:
            I = 0.
        if need_da:
            da = 0.
        if need_db:
            db = 0.
        if need_d2aa:
            d2aa = 0.
        if need_d2ab:
            d2ab = 0.
        if need_d2bb:
            d2bb = 0.
        need_first = da is not None or db is not None
        need_second = d2aa is not None or d2ab is not None or d2bb is not None
        need_deriv = need_first or need_second
        for x, w in self.X_W:
            axpb = a * x + b
            u = np.exp(axpb)
            f = 1. / (1. + u)
            if I is not None:
                I += w * f
            if need_deriv:
                u_f = np.where(axpb < 40., u * f, 1.)
                df = -f * u_f
                w_df = w * df
                if need_first:
                    if da is not None:
                        da += w_df * x
                    if db is not None:
                        db += w_df
                if need_second:
                    w_ddf = w_df * (1. - 2. * u_f)
                    if d2aa is not None:
                        d2aa += w_ddf * x * x
                    if d2ab is not None:
                        d2ab += w_ddf * x
                    if d2bb is not None:
                        d2bb += w_ddf
        return dict(value=I, da=da, db=db, d2aa=d2aa, d2ab=d2ab, d2bb=d2bb)


class GaussFermiIntegralReference(GaussFermiBase):
    """
    Reference implementation for checking accuracy, which uses adaptive
    quadrature.
    Only calculation of I is supported.
    """

    def _nevaluate(self, a, b, **kwargs):
        need_value = kwargs.pop('need_value', False)
        if any(kwargs.values()):
            raise NotImplementedError()

        def _f(x):
            return np.exp(-x * x) / (1 + np.exp(a * x + b))
        v = scipy.integrate.quad(_f, -28, 28,
                                 epsrel=1e-12, epsabs=1e-40)
        return dict(value=v[0] / np.sqrt(np.pi))


class UnivariateInterpolatedGaussFermi(GaussFermiLogBase):
    def __init__(self, impl, a, n=1000, k=5):
        super(UnivariateInterpolatedGaussFermi, self).__init__()
        self.a = a
        self.a_max = impl.a_max
        self.I_min = impl.I_min
        self.I_max = impl.I_max
        self.b_min, self.b_max = map(
            lambda b: impl.b(
                a, b), (impl.I_min, impl.I_max))
        self.btab = np.linspace(self.b_min, self.b_max, n)
        _i = impl.I(ad.forward.seed(a), self.btab)
        self.logItab = np.log(ad.nvalue(_i))
        self.dlogIdatab = 1. / ad.nvalue(_i) * _i.gradient.tocsr().todense().A1
        assert np.all(np.diff(self.logItab) > 0)
        self.spline = make_interp_spline(
            self.btab[::-1], self.logItab[::-1], k=k)
        self.daspline = make_interp_spline(
            self.btab[::-1], self.dlogIdatab[::-1], k=k)

    def _guess_b(self, a, v):
        return self.btab[np.clip(np.searchsorted(
            self.logItab, v), 0, len(self.btab) - 1)]

    def _nevaluate(self, a, b, **tasks):
        if np.any(a != self.a):
            raise ValueError('a is different than precalculated')
        if np.asarray(a).shape == (0,):
            assert np.asarray(b).shape in [(), (0,)]
            b = np.zeros(0, dtype=np.common_type(a, b))
        result = {}
        if tasks.pop('need_value', False):
            result['value'] = self.spline(b)
        if tasks.pop('need_db', False):
            result['db'] = self.spline(b, 1)
        if tasks.pop('need_d2bb', False):
            result['d2bb'] = self.spline(b, 2)
        if tasks.pop('need_da', False):
            result['da'] = self.daspline(b)
        if tasks.pop('need_d2ab', False):
            result['d2ab'] = self.daspline(b, 1)
        if any(tasks.values()):
            raise NotImplementedError()
        return result


class UnivariateInterpolatedGaussFermiFactory(object):
    def __init__(self, impl, **kwargs):
        self.I_min = impl.I_min
        self.I_max = impl.I_max
        self.a_max = impl.a_max
        self.impl = impl
        self.kwargs = kwargs
        self.d = {}

    def _get(self, a):
        a = ad.nvalue(a)
        if a.shape:
            if len(a) > 0:
                if np.count_nonzero(a != a[0]) > 0:
                    raise ValueError('a must have all values equal')
                a = a[0]
            else:
                # a is empty: return anything
                if self.d:
                    return next(iter(self.d.values()))
                else:
                    return self._get(0)
        a = float(a)
        u = self.d.get(a, None)
        if u is None:
            u = UnivariateInterpolatedGaussFermi(
                self.impl, np.asarray(a), **self.kwargs)
            self.d[a] = u
        return u

    def I(self, a, b):
        return self._get(a).I(a, b)

    def b(self, a, I):
        return self._get(a).b(a, I)

    def dIdb(self, a, b):
        return self._get(a).dIdb(a, b)


def diffusion_enhancement(nsigma_as_a, c, impl, b=None, full_output=False):
    if b is None:
        b = impl.b(nsigma_as_a, c)
    factor = -c / impl.dIdb(nsigma_as_a, b)
    if full_output:
        return dict(factor=factor, b=b)
    else:
        return factor


defaultImpl = GaussFermiIntegralH(96, 7. * np.sqrt(2.))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .source import BulkSource
from oedes.ad import exp

__all__ = ['TrapSource']


class TrapSource(BulkSource):
    "Source term fro conduction level to trap level"

    def __init__(self, transport_eq, trap_eq, fill_transport=True,
                 fill_trap=True, rrate_param=False):
        super(
            TrapSource,
            self).__init__(
            'trap',
            eq_refs=[
                'transport_eq',
                'trap_eq'])
        assert transport_eq.mesh is trap_eq.mesh, "the same mesh must be used for conduction and trap level"
        assert transport_eq.z == trap_eq.z, "inconsistent signs"
        self.transport_eq = transport_eq
        self.trap_eq = trap_eq
        self.fill_transport = fill_transport
        self.fill_trap = fill_trap
        self.rrate_param = rrate_param

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        transportenergy = ctx.param(eq.transport_eq,  'energy')
        trate = ctx.param(eq.trap_eq, 'trate')
        if self.rrate_param:
            rrate = ctx.param(eq.trap_eq, 'rrate')
        else:
            trapenergy = ctx.param(eq.trap_eq, 'energy')
            depth = (trapenergy - transportenergy) * eq.transport_eq.z
            rrate = trate * \
                exp(-depth / ctx.varsOf(eq.transport_eq.thermal)['Vt'])
        transportN0 = ctx.param(eq.transport_eq, 'N0')
        trapN0 = ctx.param(eq.trap_eq, 'N0')
        ctransport = ctx.varsOf(eq.transport_eq)['c']
        ctrap = ctx.varsOf(eq.trap_eq)['c']
        if self.fill_trap:
            a = trate * (trapN0 - ctrap)
        else:
            a = trate * trapN0
        if self.fill_transport:
            b = rrate * (transportN0 - ctransport)
        else:
            b = rrate * transportN0
        g = a * ctransport - b * ctrap
        self.add(ctx, g, plus=[eq.trap_eq], minus=[eq.transport_eq])



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['Domain']
from .equations import ConstTemperature, PoissonEquation, RamoShockleyCurrentCalculation, ChargedSpecies
from .source import BulkSource
from .compose import ModelAdapter, Coupled


class Domain(ModelAdapter, Coupled):
    "Domain in space with common Poisson's equation and thermal model"

    def __init__(self, mesh=None, calculate_current=True):
        super(Domain, self).__init__()
        self.poisson = PoissonEquation(mesh=mesh)
        self.species = []
        self.sources = []
        self.ordering = 'cell'
        self.thermal = ConstTemperature()
        self.other = []
        self.calculate_current = calculate_current
        self._current_calculation = None

    def _check(self, check, value):
        if check is not None:
            assert check is value or not check
        return value

    @property
    def parts(self):
        yield self.thermal
        assert isinstance(self.poisson, PoissonEquation)
        yield self.poisson
        for eq in self.species:
            assert isinstance(eq, ChargedSpecies)
            eq.poisson = self._check(eq.poisson, self.poisson)
            eq.thermal = self._check(eq.thermal, self.thermal)
            yield eq
        for source in self.sources:
            assert isinstance(source, BulkSource)
            yield source
        for other in self.other:
            yield other
        if self.calculate_current:
            if self._current_calculation is None:
                self._current_calculation = RamoShockleyCurrentCalculation(
                    (self.poisson,))
            yield self._current_calculation



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['OhmicContact', 'OhmicBCPoisson', 'OhmicBCSpecies']

from .boundary import AppliedVoltage, DirichletBC
import weakref
from oedes.ad import getitem
from oedes.utils import Coupled
from oedes.models.equations import WithDOS


class OhmicBCPoisson(AppliedVoltage):
    def __init__(self, name, semiconductor, poisson, **kwargs):
        super(OhmicBCPoisson, self).__init__(name, **kwargs)
        self.semiconductor = semiconductor
        self._owner_eq_weak = weakref.ref(poisson)

    def initDiscreteEq(self, builder, eq):
        super(OhmicBCPoisson, self).initDiscreteEq(builder, eq)
        eq.semiconductor = builder.get(self.semiconductor)
        eq.semiconductor.alldone.depends(eq.evaluate)
        eq.evaluate.depends(eq.semiconductor.load)

    def potential(self, ctx, eq):
        return self.voltage(
            ctx, eq) + getitem(ctx.varsOf(eq.semiconductor)['Efv'], eq._getIdx())


class OhmicBCSpecies(DirichletBC):
    def __init__(self, name, semiconductor, species):
        super(OhmicBCSpecies, self).__init__(name)
        self.semiconductor = semiconductor
        self._owner_eq_weak = weakref.ref(species)

    def initDiscreteEq(self, builder, eq):
        super(OhmicBCSpecies, self).initDiscreteEq(builder, eq)
        eq.semiconductor = builder.get(self.semiconductor)
        eq.semiconductor.alldone.depends(eq.evaluate)
        eq.evaluate.depends(eq.semiconductor.load)

    def value(self, ctx, eq):
        return getitem(ctx.varsOf(eq.semiconductor)[
                       'conc_Ef'][id(eq.owner_eq)], eq._getIdx())


class OhmicContact(Coupled):
    def __init__(self, poisson, semiconductor, name):
        parts = [OhmicBCPoisson(name, semiconductor, poisson)]
        for s in semiconductor.species:
            if isinstance(s, WithDOS):
                parts.append(OhmicBCSpecies(name, semiconductor, s))
        super(OhmicContact, self).__init__(parts)
        self.semiconductor = semiconductor
        self.poisson = poisson
        self.name = name



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import warnings
from six import BytesIO, PY2
from oedes.optical.func import FunctionOfWavelength, makeComplexFunctionOfWavelength, InterpolatedFunctionOfWavelength, ConstFunctionOfWavelength
import numpy as np
import yaml
from oedes.optical.func import Material

__all__ = ['RefractiveIndexInfoMaterial']


class RefractiveIndexInfoMaterialWarning(UserWarning):
    pass


class Formula(FunctionOfWavelength):
    def __init__(self, limits, coefficients, min_coeff=1, odd_number=1):
        if len(coefficients) < min_coeff or (
                odd_number is not None and len(coefficients) % 2 != odd_number):
            raise ValueError(
                '{name}: wrong number of coefficients'.format(
                    name=self.__class__.__name__))
        self.coefficients = coefficients
        self.limits = limits
        self.undefined = False

    def __call__(self, wavelength):
        return self.calculate(wavelength*1e6)


class Sellmeier(Formula):
    def __init__(self, limits, coefficients, variant_2=False):
        self.variant_2 = variant_2
        super(Sellmeier, self).__init__(limits, coefficients)

    def calculate(self, wavelength):
        w2 = wavelength*wavelength
        s = 1 + self.coefficients[0]
        for i in range(1, len(self.coefficients), 2):
            u = self.coefficients[i+1]
            if not self.variant_2:
                u = u*u
            s = s + self.coefficients[i]*w2 / (w2-u)
        return np.sqrt(s)


class Sellmeier_2(Sellmeier):
    def __init__(self, limits, coefficients):
        super(Sellmeier_2, self).__init__(limits, coefficients, variant_2=True)


class Cauchy(Formula):
    def calculate(self, wavelength):
        w = wavelength
        s = self.coefficients[0]
        for i in range(1, len(self.coefficients), 2):
            s = s + self.coefficients[i]*w**self.coefficients[i+1]
        return s


class Polynominal(Cauchy):
    def calculate(self, wavelength):
        return np.sqrt(super(Polynominal, self).calculate(wavelength))


class RefractiveIndexInfo(Formula):
    def __init__(self, *args):
        super(RefractiveIndexInfo, self).__init__(*args, min_coeff=9)

    def calculate(self, wavelength):
        w = wavelength
        w2 = w*w
        c = self.coefficients
        s = c[0]+(c[1]*w**c[2]/(w2-c[3]**c[4]))+(c[5]*w**c[6]/(w2-c[7]**c[8]))
        for i in range(9, len(self.coefficients), 2):
            s = s + self.coefficients[i]*w**self.coefficients[i+1]
        return np.sqrt(s)


class Gases(Formula):
    def calculate(self, wavelength):
        w = wavelength
        iw2 = 1./(w*w)
        s = 1 + self.coefficients[0]
        for i in range(1, len(self.coefficients), 2):
            s = s + self.coefficients[i]/(self.coefficients[i+1]-iw2)
        return s


class Herzberger(Formula):
    def __init__(self, limits, coefficients):
        super(
            Herzberger,
            self).__init__(
            limits,
            coefficients,
            min_coeff=4,
            odd_number=None)

    def calculate(self, wavelength):
        w2 = wavelength*wavelength
        c = self.coefficients
        u = 1./(w2-0.028)
        s = c[0]+u*(c[1]+c[2]*u)
        v = w2
        for i in range(3, len(self.coefficients)):
            s = s + c[i]*v
            v = v*v
        return s


class Retro(Formula):
    def __init__(self, limits, coefficients):
        if len(coefficients) != 4:
            raise ValueError('expecting 4 coefficients')
        super(Retro, self).__init__(limits, coefficients, odd_number=0)

    def calculate(self, wavelength):
        c = self.coefficients
        w2 = wavelength*wavelength
        s = c[0]+c[1]*w2/(w2-c[2])+c[3]*w2
        return np.sqrt((2*s+1)/(1-s))


class Exotic(Formula):
    def __init__(self, limits, coefficients):
        if len(coefficients) != 6:
            raise ValueError('expecting 6 coefficients')
        super(Exotic, self).__init__(limits, coefficients, odd_number=0)

    def calculate(self, wavelength):
        c = self.coefficients
        w = wavelength
        u = (w-c[4])
        s = c[0]+c[1]/(w*w-c[2])+(c[3]*u/(u*u+c[5]))
        return np.sqrt(s)


def gen_from_str(s):
    if not PY2 and isinstance(s, str):
        s = bytes(s, 'utf-8')
    return np.genfromtxt(BytesIO(s))


class RefractiveIndexInfoMaterial(Material):
    def _warn(self, message):
        if not self.ignore_warnings:
            warnings.warn(message, RefractiveIndexInfoMaterialWarning)

    def __init__(self, stream, name=None, ignore_warnings=False):
        self.ignore_warnings = ignore_warnings
        if isinstance(stream, str):
            with open(stream, 'rb') as f:
                yml = yaml.safe_load(f)
        else:
            yml = yaml.safe_load(stream)

        self.references = yml.get('REREFERENCES', None)
        self.comments = yml.get('COMMENTS', None)
        self.specs = yml.get('SPECS', None)

        self.n_data = None
        self.k_data = None

        for data in yml['DATA']:
            data_type = data['type'].split()
            if data_type[0] == 'formula':
                f = self._process_formula
            elif data_type[0] == 'tabulated':
                f = self._process_tabulated
            else:
                self._warn(
                    'Skipping not understood data type {data_type}'.format(
                        data_type=data['type']))
            f(data, data_type[1:])

        if self.n_data is None:
            self._warn('n data not defined')
        if self.k_data is None:
            self._warn('k data not defined')
        if self.n_data is not None and self.k_data is not None:
            self.data = makeComplexFunctionOfWavelength(
                self.n_data, self.k_data)
        else:
            self.data = None

        super(RefractiveIndexInfoMaterial, self).__init__(refractive_index=self.data, name=name)

    def _process_tabulated(self, data, args):
        rows = gen_from_str(data['data'])
        if rows.ndim == 0 or not all(rows.shape):
            self._warn('skipping empty tabulated data')
            return
        if rows.ndim == 1:
            rows = rows[np.newaxis, ...]
        rows = np.atleast_2d(rows)
        rows = rows[np.argsort(rows[:, 0])]
        delta = np.diff(rows[:, 0])
        duplicate = delta == 0.
        if np.any(duplicate):
            ix = np.ones(len(rows), dtype=np.bool)
            ix[1:][duplicate] = False
            self._warn(
                'Removing duplicate wavelength values : {dup}'.format(dup=rows[~ix, 0]))
            rows = rows[ix]
        types = {'n': ((1, None), 2), 'k': ((None, 1), 2), 'nk': ((1, 2), 3)}
        if len(args) != 1:
            raise ValueError(
                'tabulated arguments not understood {args}'.format(
                    args=args))
        type_name, = args
        if type_name not in types:
            raise ValueError(
                'tabulated type not understood {type_name}'.format(
                    type_name=type_name))
        (n_source, k_source), ncols_required = types[type_name]
        nrows, ncols = rows.shape
        if ncols != ncols_required:
            raise ValueError(
                'invalid number of columns for type {type_name} : expected {ncols_required}, got {ncols}'.format(
                    type_name=type_name, ncols_required=ncols_required, ncols=ncols))
        wavelength = rows[:, 0]*1e-6

        def to_function(x, y):
            if len(x) > 1:
                return InterpolatedFunctionOfWavelength(x, y)
            else:
                return ConstFunctionOfWavelength(y[0], (x[0], x[0]))
        if n_source is not None:
            if self.n_data is not None:
                self._warn('tabulated data redefines n data')
            self.n_data = to_function(wavelength, rows[:, n_source])
        if k_source is not None:
            if self.k_data is not None:
                self._warn('tabulated data redefines k data')
            self.k_data = to_function(wavelength, rows[:, k_source])

    def _process_formula(self, data, args):
        if self.n_data is not None:
            self._warn('formula is redefining n data')
        if len(args) != 1:
            raise ValueError(
                'formula arguments not understood {args}'.format(args))
        formula_type, = args
        formula_types = {
            '1': Sellmeier,
            '2': Sellmeier_2,
            '3': Polynominal,
            '4': RefractiveIndexInfo,
            '5': Cauchy,
            '6': Gases,
            '7': Herzberger,
            '8': Retro,
            '9': Exotic
        }
        if formula_type not in formula_types:
            raise ValueError(
                'unknown formula type {formula_type}'.format(
                    formula_type=formula_type))
        coefficients = gen_from_str(data['coefficients'])
        limits = None
        if 'range' in data:
            limits = data['range']
        if 'wavelength_range' in data:
            if limits is not None:
                raise ValueError(
                    'cannot specify range and wavelength_range at once')
            limits = data['wavelength_range']
        limits = gen_from_str(limits)*1e-6
        if limits.shape != (2,) or limits[0] > limits[1]:
            raise ValueError(
                'wavelength limits not understood {limits}'.format(
                    limits=limits))
        if self.n_data is not None:
            raise ValueError('redefining n')
        self.n_data = formula_types[formula_type](limits, coefficients)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['Electroneutrality']

import numpy as np
import scipy.optimize
from oedes.utils import Calculation
from oedes.models import solver
from oedes.ad import exp, log, sqrt, nvalue, custom_function, forward, isscalar
from .dos import WithDOS
from .boltzmann import BoltzmannDOS
from oedes.functions import brent


def _same(seq):
    value = next(seq)
    if all([x is value for x in seq]):
        return value
    else:
        return None


class Electroneutrality(Calculation):
    "This finds Fermi level corresponding to electrical neutrality, as well as neutral charge concentrations"

    etol = 1e-15
    force_general = False

    def __init__(self, species, name=None):
        super(Electroneutrality, self).__init__(name)
        self.species = species
        with_dos = []
        other = []
        for s in self.species:
            if isinstance(s, WithDOS):
                with_dos.append(s)
            else:
                other.append(s)
        self.with_dos = with_dos
        self.other = other
        self.electron = [eq for eq in with_dos if eq.z == -1]
        self.hole = [eq for eq in with_dos if eq.z == 1]

    def initDiscreteEq(self, builder, eq):
        super(Electroneutrality, self).initDiscreteEq(builder, eq)
        with_dos = list(map(builder.get, self.with_dos))
        other = list(map(builder.get, self.other))
        species = with_dos + other
        eq.with_dos, eq.other, eq.species = with_dos, other, species
        eq.electron = list(map(builder.get, self.electron))
        eq.hole = list(map(builder.get, self.hole))
        for s in species:
            s.alldone.depends(eq.load)

        # check if same mesh and same poisson equation
        eq.mesh = _same(s.mesh for s in species)
        eq.poisson = _same(s.poisson for s in species)
        eq.load.depends(eq.poisson.load)
        for s in with_dos:
            eq.load.depends(s.thermal.load)

        # Check conditions for application of analytical solution:
        # - bandgap, two species, different charges +-e
        # - both use Boltzann DOS
        # - both use same effective temperature
        same_thermal = _same(s.thermal for s in with_dos)
        boltzmann = _same(isinstance(s.dos, BoltzmannDOS) for s in with_dos)
        bandgap = (self.electron and eq.hole and len(
            with_dos) == 2) or len(with_dos) == 1
        eq.analytical = same_thermal and boltzmann and bandgap
        if eq.analytical:
            eq.mesh = eq.with_dos[0].mesh
            eq.thermal = eq.with_dos[0].thermal

    def solve_analytical(self, ctx, eq, fixed):
        hole, = eq.hole
        electron, = eq.electron
        C = fixed
        Nc = electron.dos.N0(ctx, electron)
        Nv = hole.dos.N0(ctx, hole)
        Vt = ctx.varsOf(eq.thermal)['Vt']
        Ec = ctx.varsOf(electron)['Ebandv']
        Ev = ctx.varsOf(hole)['Ebandv']
        uc = exp(Ec/Vt)
        uv = exp(Ev/Vt)
        # TODO: rewrite to avoid overflow, reference to middle of bandgap
        Ef = Vt*log((C*uc + sqrt((C*C*uc + 4*Nc*Nv*uv)*uc))/(2*Nc))
        n = Nc*exp((Ef-Ec)/Vt)
        p = Nv*exp((Ev-Ef)/Vt)
        return Ef, dict({id(hole): p, id(electron): n})

    def solve_general(self, ctx, eq, fixed):
        def conc(Ef, numeric_parameters):
            for s in eq.with_dos:
                yield s, s.dos.concentrationv(ctx, s, None, Ef, numeric_parameters=numeric_parameters)
        nfixed = nvalue(fixed)

        def f(Ef, numeric_parameters=True):
            if numeric_parameters:
                charge = nfixed
            else:
                charge = fixed
            for eq, c in conc(Ef, numeric_parameters):
                charge = charge + c*eq.z
            return charge
        bandv = np.asarray([nvalue(ctx.varsOf(s)['Ebandv'])
                            for s in eq.with_dos])
        a = np.amax(bandv)
        b = np.amin(bandv)
        Ef_value = brent(f, a, b, xtol=self.etol, maxiter=50)
        g = f(Ef_value, False)
        if isinstance(g, forward.value):
            raise NotImplementedError('not tested')
            dg = f(forward.seed(Ef_value), True).deriv
            if not isscalar(dg):
                dg = dg.tocsr().diagonal()
            # must create value : (Ef_value, 1/dg*g')
            Ef = custom_function(lambda *args: Ef_value, lambda *args: dg)(g)
        else:
            Ef = Ef_value
        info = dict((id(eq), c) for eq, c in conc(Ef, False))
        return Ef, info

    def load(self, ctx, eq):
        super(Electroneutrality, self).load(ctx, eq)
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            return
        fixed = 0
        for s in eq.other:
            fixed = fixed + nvalue(ctx.varsOf(s)['c']) * s.z
        if eq.analytical and not self.force_general:
            Efv, c = self.solve_analytical(ctx, eq, fixed)
        else:
            Efv, c = self.solve_general(ctx, eq, fixed)
        ctx.varsOf(eq)['Efv'] = Efv
        ctx.varsOf(eq)['conc_Ef'] = c
        Ef = -ctx.varsOf(eq.poisson)['potential'] + Efv
        ctx.outputCell([eq, 'Ef'], Ef, unit=ctx.units.eV)
        ctx.outputCell([eq, 'phi'], -Ef, unit=ctx.units.V)
        for s, sc in zip(eq.with_dos, self.with_dos):
            ctx.outputCell([eq, sc.name, 'c'], c[id(s)],
                           unit=ctx.units.concentration)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from oedes import logs


class ConvergenceTest(object):
    pass


class ElementwiseConvergenceTest(ConvergenceTest):

    def __init__(self, rtol=1e-6, atol=0., eps=0.):
        self.rtol = rtol
        self.atol = atol
        self.eps = eps

    def _testix(self, eq, ix, F, x, dx, report, prefix):
        i = eq.idx[ix]
        if len(i) == 0:
            return True
        F = F[i]
        x = x[i]
        dx = dx[i]
        converged_abs = np.abs(F) < self.atol
        converged_rel = np.abs(dx) < (self.rtol * np.abs(x) + self.eps)
        converged = np.logical_or(converged_abs, converged_rel).all()

        def amax(v):
            if len(v) == 0:
                return 0.
            else:
                return np.amax(v)
        with np.errstate(divide='ignore', invalid='ignore'):
            infodict = dict(converged=converged,
                            atol=self.atol,
                            rtol=self.rtol,
                            maxabs=amax(np.abs(F)),
                            maxabs_nc=amax(
                                np.abs(F[np.logical_not(converged_rel)])),
                            maxrel=amax(np.abs(dx / x)),
                            maxrel_nc=amax(np.abs((dx / x)[np.logical_not(converged_abs)])))
            report[prefix] = infodict
        logger = logs.nonlinear_convergence
        if logger.isEnabledFor(logs.logging.DEBUG):
            logger.debug('%s on %s:' % (self.__class__.__name__, prefix))

            def fmt(item):
                k, v = item
                if isinstance(v, float) or isinstance(v, np.ndarray):
                    sv = '%.3e' % v
                else:
                    sv = str(v)
                return '%s=%s' % (k, sv)
            logger.debug(' %s' % ' '.join(list(map(fmt, infodict.items()))))
        return converged

    def testEquation(self, eq, F, x, dx, report):
        ix_all = np.hstack(
            [eq.mesh.internal.idx, eq.mesh.boundary.idx[eq.bc_dof_is_free]])
        return self._testix(eq, ix_all, F, x, dx, report, eq.prefix)

    def testBoundary(self, eq, bc, F, x, dx, report):
        assert bc.owner_eq is eq
        return self._testix(eq, eq.mesh.boundary.idx[
                            bc.boundary['bidx']], F, x, dx, report, '.'.join([str(eq.prefix), bc.name]))


class DummyConvergenceTest(ConvergenceTest):
    def testEquation(self, eq, F, x, dx, report):
        return True

    def testBoundary(self, eq, bc, F, x, dx, report):
        return True


class DiscreteEquation(object):
    def __init__(self, name=None):
        self.convergenceTest = None
        self.idx = None
        self.transientvar = 1.
        self.name = name

    def init(self, builder):
        pass

    def update(self, x, *args):
        pass

    def scaling(self, xscaling, fscaling):
        pass

    def testConverged(self, F, x, dx, report):
        return self.convergenceTest.testEquation(self, F, x, dx, report)


class GeneralDiscreteEquation(DiscreteEquation):
    def __init__(self, ndof, name=None):
        super(GeneralDiscreteEquation, self).__init__(name=name)
        self._ndof = ndof

    def ndof(self):
        return self._ndof


class DelegateConvergenceTest(object):
    def __init__(self, target):
        self.target = target

    def testEquation(self, eq, F, x, dx, report):
        return self.target.convergenceTest.testEquation(
            eq, F, x, dx, report)

    def testBoundary(self, eq, bc, F, x, dx, report):
        return self.target.convergenceTest.testBoundary(
            eq, bc, F, x, dx, report)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from sparsegrad import *
from sparsegrad.base import *
from sparsegrad import forward
from sparsegrad.forward import nvalue
from sparsegrad.sparsevec import *


def isscalar(x):
    return not forward.nvalue(x).shape


def getitem(x, idx):
    if idx is None:
        return x
    if not isscalar(idx) and not len(idx):
        return np.zeros(0, dtype=nvalue(x).dtype)
    if isscalar(x):
        return x
    else:
        return x[idx]


sparsesum = sparsesum_bare


def custom_function(f, df):
    f = expr.wrapped_func(func.custom_func(f, df))

    def _(*args):
        return f(*args)
    return f



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .source import BulkSource

__all__ = ['DirectGeneration', 'SimpleGenerationTerm', 'SimpleDecayTerm']


class DirectGeneration(BulkSource):
    def __init__(self, eqs, function, name='generation', param_name='I'):
        super(
            DirectGeneration,
            self).__init__(
            name=name,
            eq_reflists=['eqs'])
        assert sum(eq.z for eq in eqs) == 0
        self.function = function
        self.eqs = eqs
        self.param_name = param_name

    def initDiscreteEq(self, builder, eq):
        super(DirectGeneration, self).initDiscreteEq(builder, eq)
        eq.eqs = tuple(map(builder.get, self.eqs))
        meshes = tuple(map(lambda eq: eq.mesh, eq.eqs))
        assert all([m == meshes[0] for m in meshes[1:]])

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        G = self.function(eq.eqs[0].mesh.cells['center'])
        g = G * ctx.param(eq, self.param_name)
        ctx.output([eq, 'G'], g)
        self.add(ctx, g, plus=eq.eqs)


class SimpleGenerationTerm(DirectGeneration):
    def __init__(self, eq, function, name='generation'):
        super(
            SimpleGenerationTerm,
            self).__init__(
            [eq],
            function,
            name=eq.name +
            '.' +
            name)


class SimpleDecayTerm(BulkSource):
    def __init__(self, eq, name='decay'):
        super(SimpleDecayTerm, self).__init__(name=name, eq_refs=['eq'])
        assert eq.z == 0
        self.eq = eq

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        f = ctx.varsOf(eq.eq)['c'] * ctx.param(eq.eq, self.name)
        self.add(ctx, f, minus=[eq.eq])



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
import sys
import time

oedes = logging.getLogger('oedes')
solver = logging.getLogger('oedes.solver')
timestepping = logging.getLogger('oedes.solver.timestepping')
nonlinear = logging.getLogger('oedes.solver.nonlinear')
nonlinear_convergence = logging.getLogger('oedes.solver.nonlinear.convergence')
linear = logging.getLogger('oedes.solver.linear')
models = logging.getLogger('oedes.models')
fvm = logging.getLogger('oedes.fvm')


class IndentedFormatter(logging.Formatter):
    def _indentLevel(self, name):
        if name.startswith('oedes.solver.timestepping'):
            return 0
        if name.startswith('oedes.solver.nonlinear'):
            if name.startswith('oedes.solver.nonlinear.convergence'):
                return 2
            if name.startswith('oedes.solver.nonlinear.timing'):
                return 2
            return 1
        if name.startswith('oedes.solver.linear'):
            return 2
        if name.startswith('oedes.models'):
            return 2
        if name.startswith('oedes.fvm'):
            return 2
        return 0

    def format(self, record):
        return '\t' * self._indentLevel(record.name) + record.msg


class Timer(object):
    def __init__(self):
        self.t0 = time.clock()
        self.t1 = None

    def elapsed(self):
        if self.t1 is None:
            self.stop()
        return self.t1 - self.t0

    def start(self):
        self.t0 = time.clock()

    def stop(self):
        self.t1 = time.clock()

    def __str__(self):
        return '%.3e s' % self.elapsed()


class LogOutput(object):
    def __init__(self, level, handler=None, logger=None):
        if handler is None:
            handler = logging.StreamHandler()
            handler.setLevel(level)
        if logger is None:
            logger = oedes
        self.handler = handler
        self.logger = logger
        self.level = level
        self.formatter = IndentedFormatter()

    def __enter__(self):
        self.handler.setFormatter(self.formatter)
        self.logger.addHandler(self.handler)

    def __exit__(self, *args):
        self.logger.removeHandler(self.handler)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from setuptools import setup, Command

with open('oedes/_version.py') as version_file:
    exec(version_file.read())


class build_doc(Command):
    description = "build the documentation"

    user_options = []

    def run(self):
        import sphinx
        sphinx.main(['sphinx-build', '-b', 'html', 'doc', 'doc/_build/html'])

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class run_notebooks(Command):
    description = 'run notebooks'

    user_options = []

    def run(self):
        import fnmatch
        import nbformat
        from nbconvert.preprocessors import ExecutePreprocessor
        import os.path
        import os
        all_ok = True
        for root, dirnames, filenames in os.walk('.'):
            if os.path.split(root)[-1] == '.ipynb_checkpoints':
                continue
            for name in fnmatch.filter(filenames, '*.ipynb'):
                f = os.path.join(root, name)
                print(f)
                try:
                    nb = nbformat.read(f, as_version=nbformat.NO_CONVERT)
                    ep = ExecutePreprocessor(timeout=600)
                    ep.preprocess(
                        nb, {'metadata': {'path': os.path.abspath(os.path.split(f)[0])}})
                    del ep
                except BaseException:
                    print('failed to execute %s' % f)
                    all_ok = False
                    continue
                nbformat.write(nb, f)
        assert all_ok

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


setup(name='oedes',
      version=version,
      packages=['oedes',
                'oedes.models',
                'oedes.models.tests',
                'oedes.models.equations',
                'oedes.models.source',
                'oedes.functions',
                'oedes.functions.tests',
                'oedes.fvm',
                'oedes.optical',
                'oedes.optical.databases',
                'oedes.optical.test_materials',
                'oedes.optical.spectra',
                'oedes.optical.tests',
                'oedes.tests',
                'oedes.mpl',
                'oedes.utils'
                ],
      package_data={'oedes.optical.test_materials': ['*.yml'],
                    'oedes.optical.spectra': ['*.zip']},
      url='https://oedes.org/',
      license='GNU Affero General Public License v3',
      author='Marek Zdzislaw Szymanski',
      install_requires=[
          'numpy',
          'scipy',
          'sparsegrad>=0.0.9',
          'tqdm>=4.19',
          'nose',
          'matplotlib',
          'pyyaml',
          'six'
      ],
      author_email='marek@marekszymanski.com',
      description='organic electronic device simulator',
      long_description=open('README.rst').read(),
      include_package_data=True,
      platforms='any',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Development Status :: 3 - Alpha',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: GNU Affero General Public License v3',
          'Topic :: Scientific/Engineering :: Mathematics',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)'
      ],
      cmdclass={
          'build_doc': build_doc,
          'run_notebooks': run_notebooks
      }
      )



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.utils import Calculation
from oedes import functions
from oedes.models.equations import TransportedSpecies
from oedes.ad import sum

__all__ = [
    'BulkSource',
    'LangevinRecombination',
    'DirectRecombination',
    'OutcouplingCalculation']


class BulkSource(Calculation):
    def __init__(self, name, eq_refs=[], eq_reflists=[]):
        super(BulkSource, self).__init__(name=name)
        self._eq_refs = eq_refs
        self._eq_reflists = eq_reflists

    def add(self, ctx, r, plus=[], minus=[]):
        z = 0
        for p in plus:
            pvars = ctx.varsOf(p)
            pvars['sources'] = pvars['sources'] + r
            z = z + p.z
        for n in minus:
            nvars = ctx.varsOf(n)
            nvars['sources'] = nvars['sources'] - r
            z = z - n.z
        assert z == 0, 'not charge neutral'

    def initDiscreteEq(self, builder, obj):
        super(BulkSource, self).initDiscreteEq(builder, obj)

        for k in self._eq_refs:
            eq = builder.get(getattr(self, k))
            obj.evaluate.depends(eq.load)
            eq.evaluate.depends(obj.evaluate)
            setattr(obj, k, eq)
        for k in self._eq_reflists:
            eqs = tuple(map(builder.get, getattr(self, k)))
            for eq in eqs:
                obj.evaluate.depends(eq.load)
                eq.evaluate.depends(obj.evaluate)
            setattr(obj, k, eqs)


class _Recombination(BulkSource):
    "Langevinian recombination term, to be used"

    def __init__(self, *args, **kwargs):
        output_name = kwargs.pop('output_name', 'R')
        name = kwargs.pop('name', None)
        self.skip_intrinsic = kwargs.pop('skip_intrinsic', False)
        if len(args) == 2:
            electron_eq, hole_eq = args
            self.semiconductor = None
        else:
            self.semiconductor, = args
            electron_eq, = self.semiconductor.electron
            hole_eq, = self.semiconductor.hole

        super(
            _Recombination,
            self).__init__(
            name,
            eq_refs=[
                'electron_eq',
                'hole_eq'], **kwargs)
        assert electron_eq.z == -1
        assert hole_eq.z == 1
        assert hole_eq.poisson is electron_eq.poisson

        self.electron_eq = electron_eq
        self.hole_eq = hole_eq
        self.output_name = output_name

    def initDiscreteEq(self, builder, eq):
        super(_Recombination, self).initDiscreteEq(builder, eq)
        if self.semiconductor is not None:
            eq.semiconductor = builder.get(self.semiconductor)
            eq.semiconductor.alldone.depends(eq.evaluate)
            eq.evaluate.depends(eq.semiconductor.load)
        else:
            eq.semiconductor = None
        eq.mesh = eq.electron_eq.mesh

    def intrinsic_from_semiconductor(self, ctx, eq):
        svars = ctx.varsOf(eq.semiconductor)
        return svars['conc_Ef'][id(eq.hole_eq)] * \
            svars['conc_Ef'][id(eq.electron_eq)]

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        R = self.evaluate_recombination(ctx, eq)
        ctx.varsOf(eq)['R'] = R
        ctx.outputCell([eq,
                        self.output_name],
                       R,
                       mesh=eq.mesh,
                       unit=ctx.units.dconcentration_dt)
        self.add(ctx, R, minus=[eq.electron_eq, eq.hole_eq])

    def intrinsic_from_params(self, ctx, eq):
        return ctx.common_param([eq.electron_eq, eq.hole_eq], 'npi')

    def intrinsic(self, ctx, eq):
        if self.skip_intrinsic:
            return 0.
        if eq.semiconductor is not None:
            return self.intrinsic_from_semiconductor(ctx, eq)
        else:
            return self.intrinsic_from_params(ctx, eq)


class LangevinRecombination(_Recombination):
    def evaluate_recombination(self, ctx, eq):
        nvars = ctx.varsOf(eq.electron_eq)
        pvars = ctx.varsOf(eq.hole_eq)
        if not isinstance(self.electron_eq, TransportedSpecies):
            nvars = dict(c=nvars['c'], mu_cell=0.)
        if not isinstance(self.hole_eq, TransportedSpecies):
            pvars = dict(c=pvars['c'], mu_cell=0.)
        epsilon = ctx.varsOf(eq.electron_eq.poisson)['epsilon']
        return functions.LangevinRecombination(nvars['mu_cell'], pvars['mu_cell'], nvars['c'], pvars['c'],
                                               epsilon, self.intrinsic(ctx, eq))


class DirectRecombination(_Recombination):
    def __init__(self, *args, **kwargs):
        self.param_name = kwargs.pop('param_name', 'beta')
        super(DirectRecombination, self).__init__(*args, **kwargs)

    def evaluate_recombination(self, ctx, eq):
        return ctx.param(eq, self.param_name) * \
            (ctx.varsOf(eq.electron_eq)['c'] *
             ctx.varsOf(eq.hole_eq)['c'] -
             self.intrinsic(ctx, eq))


class OutcouplingCalculation(Calculation):
    def __init__(self, recombination_eqs, name='outcoupling', output_name='L'):
        super(OutcouplingCalculation, self).__init__(name)
        self.recombination_eqs = recombination_eqs
        self.output_name = output_name

    def initDiscreteEq(self, builder, obj):
        super(OutcouplingCalculation, self).initDiscreteEq(builder, obj)
        obj.recombination_eqs = tuple(map(builder.get, self.recombination_eqs))
        for eq in obj.recombination_eqs:
            obj.evaluate.depends(eq.evaluate)

    def evaluate(self, ctx, eq):
        if not ctx.wants_output:
            return
        if ctx.solver.poissonOnly:
            return
        total = 0
        for recombination_eq in eq.recombination_eqs:
            total = total + \
                sum(recombination_eq.mesh.cells['center']
                    * ctx.varsOf(recombination_eq)['R'])
        ctx.output([eq, self.output_name], total)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .computation import Computation
from .funcall import Funcall

_all__ = ['Equation', 'Calculation', 'EquationWithMesh', 'SubEquation']


class Equation(Computation):
    def __init__(self, name=None):
        super(Equation, self).__init__()

        self.name = name
        self.convergenceTest = None

    def all_equations(self, args):
        yield (args.push(name=self.name), self)

    def newDiscreteEq(self, builder):
        return builder.newGeneralDiscreteEquation(0)

    def initDiscreteEq(self, builder, obj):
        "Late initialization of discrete equation object (after all objects were created)"
        pass

    def build(self, builder, args):
        eq = self.newDiscreteEq(builder)
        eq.prefix = args.prefix
        self.buildDiscreteEq(builder, eq)
        builder.addInitializer(self.initDiscreteEq, eq)
        return eq

    def load(self, ctx, eq):
        vars = ctx.newVars(eq)
        vars['x'] = ctx.gvars['x'][eq.idx]
        vars['xt'] = ctx.gvars['xt'][eq.idx]

    def evaluate(self, ctx, eq):
        pass

    def buildDiscreteEq(self, builder, obj):
        obj.convergenceTest = builder.newDelegateConvergenceTest(self)
        obj.load = Funcall(self.load, obj)
        obj.evaluate = Funcall(self.evaluate, obj, depends=[obj.load])
        obj.alldone = Funcall(None, obj, depends=[obj.evaluate])
        builder.addEvaluation(obj.load)
        builder.addEvaluation(obj.evaluate)
        builder.addEvaluation(obj.alldone)
        builder.add(self, obj)


class SubEquation(object):
    def subInit(self, parent, builder, parent_eq):
        pass


class EquationWithMesh(Equation):
    def __init__(self, mesh, name=None):
        super(EquationWithMesh, self).__init__(name)
        self.mesh = mesh


class Calculation(Equation):
    def buildDiscreteEq(self, builder, obj):
        super(Calculation, self).buildDiscreteEq(builder, obj)
        obj.convergenceTest = builder.newDelegateConvergenceTest(None)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['brent']

import numpy as np


def sort(a, b, fa, fb):
    "if |f(a)|<|f(b)| swap(a,b)"
    swap = np.abs(fa) < np.abs(fb)

    def aswap(s, a, b):
        a_ = np.where(s, b, a)
        b_ = np.where(s, a, b)
        return a_, b_
    a, b = aswap(swap, a, b)
    fa, fb = aswap(swap, fa, fb)
    return a, b, fa, fb


def converged(a, b, s, fa, fb, fs, xtol):
    cb = fb == 0.
    if s is not None:
        cs = fs == 0.
    else:
        cs = False
    ctol = abs(a-b) < 2.*xtol
    if not np.all(np.logical_or(np.logical_or(cb, cs), ctol)):
        return None
    x = (a+b)/2
    x = np.where(cb, b, x)
    if s is not None:
        x = np.where(cs, s, x)
    return x


def quadratic(a, b, c, fa, fb, fc):
    with np.errstate(divide='ignore', invalid='ignore'):
        dab = fa-fb
        dbc = fb-fc
        dca = fc-fa
        x = -a*fb*fc/(dab*dca)-b*fa*fc/(dab*dbc)-c*fa*fb/(dca*dbc)
        return x


def secant(a, b, fa, fb):
    return b-fb*(b-a)/(fb-fa)


def in_interval(x, a, b):
    return np.logical_and(x >= np.minimum(a, b), x <= np.maximum(a, b))


def brent(f, a, b, xtol, maxiter=10):
    "Solve f(x) = 0, with a,b "
    fa = f(a)
    fb = f(b)
    assert np.all(fa*fb < 0), 'function not bracketed'
    a, b, fa, fb = sort(a, b, fa, fb)
    c, fc = a, fa
    mflag, d, s, fs = True, 0., None, None
    for iteration in range(maxiter):
        assert (fa == f(a)).all() and (fb == f(b)).all() and (fc == f(c)).all()
        assert s is None or (fs == f(s)).all()
        assert np.all(fa*fb <= 0), 'lost bracket'
        assert (np.abs(fa) >= np.abs(fb)).all()
        assert (np.abs(fa) >= np.abs(fc)).all()
        assert (np.abs(fc) >= np.abs(fb)).all()
        cq = converged(a, b, s, fa, fb, fs, xtol)
        if cq is not None:
            return cq
        s = quadratic(a, b, c, fa, fb, fc)
        s = np.where(np.isfinite(s), s, secant(a, b, fa, fb))
        u = np.where(mflag, np.abs(b-c), np.abs(c-d))
        cond1 = np.logical_not(in_interval(s, (3*a+b)/4, b))
        cond2_3 = abs(s-b) >= u/2
        cond4_5 = u < xtol
        mflag = cond1 | cond2_3 | cond4_5
        s = np.where(mflag, (a+b)*0.5, s)
        assert np.all(np.isfinite(s))
        fs = f(s)
        d, c = c, b
        fc = fb
        cond = fa*fs < 0
        b = np.where(cond, s, b)
        fb = np.where(cond, fs, fb)
        a = np.where(cond, a, s)
        fa = np.where(cond, fa, fs)
        a, b, fa, fb = sort(a, b, fa, fb)
    raise RuntimeError('iteration limit exceeded')



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# # Transport of charged species
#
# $v=z \mu E$
#
# $D=\mu V_t$
#
# $V_t=\frac{k_B T}{q}$
#
# Fluxes for electrons and holes are
#
# $j_n=-\mu_e (n E + V_t \nabla n)$
#
# $j_p=\mu_h (p E - V_t \nabla p)$
#
# Electric currents are
#
# $J_n=q \mu_e (n E + V_t \nabla n)$
#
# $J_p=q \mu_h (p E - V_t \nabla p)$

from .cell import FVMConservationEquation
from oedes.functions import ScharfetterGummelFlux
import scipy.constants
import numpy as np
from oedes import logs


class FVMTransportEquation(FVMConservationEquation):
    "Advection-diffusion equation discretized with Scharfetter-Gummel scheme"

    def faceflux(self, c, v, D, full_output=False):
        """
        Return Scharfetter-Gummel approximation of flux

        Parameters:
        -----------
        c: vector
            concentration per cell
        v: vector
            velocity per face
        D: vector
            diffusion term per face
        full_output: bool
            If True, returns dict (flux,cmid,cgrad,flux_v,flux_d), otherwise flux.

        Returns:
        --------
        flux:
            Flux per face
        cmid:
            Concetration per face
        cgrad:
            Concetration gradient per face
        flux_v:
            Advection term of flux per face
        flux_d:
            Diffusion term of flux per face
        """
        dr = self.mesh.faces['dr']
        cfrom = c[self.mesh.faces['-']]
        cto = c[self.mesh.faces['+']]
        if not full_output:
            return ScharfetterGummelFlux(
                dr, cfrom, cto, v, D, full_output=False)
        flux, cmid, cgrad = ScharfetterGummelFlux(
            dr, cfrom, cto, v, D, full_output=True)
        return dict(flux=flux, cmid=cmid, cgrad=cgrad,
                    flux_v=cmid * v, flux_D=-D * cgrad)

    def update(self, x, *args):
        # Do not allow negative values
        xe = np.take(x, self.idx)
        xnew = np.where(xe < 0., 0., xe)
        if np.any(xe != xnew):
            pass
            #logs.fvm.info('update(%r): clipping concentration c<0, min(c)=%r'%(self, np.amin(xe)))
        x[self.idx] = xnew

    def scaling(self, xscaling, fscaling):
        lunit = 1e-9
        tunit = 1e-12
        xscaling[self.idx] = lunit**-3
        fscaling[self.idx] = lunit * tunit


class FVMTransportChargedEquation(FVMTransportEquation):
    "Advection-diffusion equation for charged species"

    def __init__(self, mesh, name, z):
        super(FVMTransportChargedEquation, self).__init__(mesh, name)
        self.z = z
        self.ze = z * scipy.constants.elementary_charge

    def v(self, mu, E):
        return E * mu * self.z

    def D(self, mu, D_mu):
        return mu * D_mu

    def chargedensity(self, c):
        return c * self.ze

    def facecurrent(self, faceflux):
        return faceflux * self.ze



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import solver
from .model import model
from .models import Poisson
from .models.solver import PoissonOnly, SolverObject

from collections import namedtuple
import numpy as np
import bisect

contextstate = namedtuple('contextstate', ['x', 'xt', 'time', 'params'])

import tqdm

progressbar_type = tqdm.tqdm


class solution_vector(object):
    "Solution vector with metadata: model, time, params, and references to solutions in previous timesteps for transient simulation"

    def __init__(self, model, time, x, params, old=(), weights=None):
        self.model = model
        self.time = time
        self.x = x
        self.old = old
        self.params = params
        self.weights = weights

    def xt(self):
        "Numerically evaluate xt using weights"
        return self.weights[0] * self.x + \
            sum(o.x * w for o, w in zip(self.old, self.weights[1:]))

    def interpolate(self, time):
        "Return interpolated solution vector"
        raise NotImplementedError()


class steadystate_solution(solution_vector):
    "Steady-state solution"

    def __init__(self, model, time, x, params):
        super(
            steadystate_solution,
            self).__init__(
            model,
            time,
            x,
            params,
            old=(),
            weights=(
                0.,
            ))


class transient_solution(solution_vector):
    "Transient solution"
    pass


class bdf_solution(transient_solution):
    def __init__(self, model, time, x, params, old, weights):
        super(
            bdf_solution,
            self).__init__(
            model,
            time,
            x,
            params,
            old=old,
            weights=weights)
        assert len(old) + 1 == len(weights)


class bdf1_solution(bdf_solution):
    def __init__(self, model, time, x, params, old, dt):
        super(
            bdf1_solution, self).__init__(
            model, time, x, params, old, weights=[
                1. / dt, -1. / dt])

    def interpolate(self, time):
        ts0 = self.old[0]
        ts1 = self

        assert ts0.time <= time
        assert ts1.time >= time
        k = (time - ts0.time) / (ts1.time - ts0.time)
        k1 = k
        k0 = (1. - k)
        x = ts0.x * k0 + ts1.x * k1
        #xt=ts0.xt() + ts1.xt() * k1
        xt = ts1.xt()
        params = ts1.params
        return const_solution_vector(
            self.model, time=time, x=x, xt=xt, params=params)


class const_solution_vector(solution_vector):
    "Externally provided vectors, which may not solve the model"

    def __init__(self, model, time=0., x=None, xt=None, params=None):
        if x is None:
            x = np.zeros_like(model.X)
        if xt is None:
            xt = np.zeros_like(model.X)
        if params is None:
            params = {}
        super(const_solution_vector, self).__init__(model, time, x, params, ())
        self._xt = xt

    def xt(self):
        return self._xt


class context(object):
    """
    context is top-level object which should be used to run simulations, instead of calling solver and model routines directly.
    context keeps current solution, and optionally, a sequence of obtained solutions (trajectory). It provides easy to use solver, postprocessing, and plotting routines.
    """

    def __init__(self, obj, **kwargs):
        "context can be created as a copy of another context, or by providing model and optionally solution details as keyword arguments"
        self.trajectory = []
        self._output = None
        self.add_hooks = []

        if isinstance(obj, context):
            self.model = obj.model
            self.current = obj.solution
            if kwargs:
                raise ValueError(
                    'keyword arguments not supported when building context from context')
        elif isinstance(obj, model):
            self.model = obj
            self.current = const_solution_vector(obj, **kwargs)
        elif isinstance(obj, solution_vector):
            self.model = obj.model
            self.current = obj
            if kwargs:
                raise ValueError('arguments not supported')
        else:
            raise ValueError('must provide model or context as first argument')
        self._output = None

    # Basic output
    @property
    def x(self):
        return self.current.x

    @property
    def params(self):
        return self.current.params

    @property
    def time(self):
        return self.current.time

    @property
    def xt(self):
        return self.current.xt()

    def _get_solution(self):
        return self.current

    def _set_solution(self, solution):
        self.current = solution
        self._output = None

    solution = property(_get_solution, _set_solution)

    def output(self):
        "Provide output for current solution"
        if self._output is None:
            self._output = self.model.output(
                self.time, self.x, self.xt, self.params, solver=SolverObject())
        return self._output

    # Plotting
    @property
    def pylab(self):
        import oedes.mpl as mpl
        return mpl.forcontext(self)

    def mpl(self, *args, **kwargs):
        import oedes.mpl as mpl
        return mpl.forcontext(self, *args, **kwargs)

    # Interpolating transient simulation
    def _find(self, time):
        i = 0
        if not self.trajectory:
            raise ValueError('not trajectory available')

        def _bsearch(a, cmp):
            i, j, k = 0, len(a), -1
            while i < j:
                k = (i + j) // 2
                c = cmp(a[k])
                if c == 0:
                    break
                elif c < 0:
                    i = k + 1
                else:
                    j = k
            if cmp(a[k]) > 0:
                return k - 1
            return k
        i = _bsearch(self.trajectory, lambda ts: ts.time - time)
        if i + 1 >= len(self.trajectory):
            if self.trajectory[-1].time == time:
                i = len(self.trajectory) - 2
            else:
                raise ValueError('time after last trajectory point')
        return i

    def attime(self, time):
        i = self._find(time)
        return context(self.trajectory[i + 1].interpolate(time))

    # Evaluating functions
    def _eval(self, f):
        if callable(f):
            return f(self)
        elif f == 'time':
            return self.time
        elif f in self.params:
            return self.params[f]
        else:
            output = self.output()
            return self.output()[f]

    def eval(self, *functions):
        return tuple(map(self._eval, functions))

    # Evaluations on trajectory
    def timesteps(self):
        for ts in self.trajectory:
            yield context(ts)

    steps = timesteps

    def teval(self, *functions):
        trajectory = list(self.timesteps())

        def teval_function(f):
            return np.asarray([t._eval(f) for t in trajectory])
        return tuple(map(teval_function, functions))

    def teval_dict(self, *functions):
        values = self.teval(*functions)
        return dict((f, v) for f, v in zip(functions, values))

    # Solving
    def _psolve(self, params, xguess, kwargs):
        kwargs = kwargs.copy()
        kwargs['solver'] = PoissonOnly()
        return solver.solve(self.model, xguess, params, niter=1, **kwargs)

    def _solve(self, xguess, params, pseudo_tmax=1e6, pseudo_dt0=1e-9,
               pseudo_mindt=1e-20, pseudo_maxsteps=200, use_poisson_guess=True, allow_transient=True, **kwargs):
        if use_poisson_guess:
            xguess = self._psolve(params, xguess=xguess, kwargs=kwargs)
        try:
            return solver.solve(self.model, xguess, params, **kwargs)
        except solver.SolverError:
            if not allow_transient:
                raise
            for t, dt, x, xt, out in solver.bdf1adapt_(
                    self.model, xguess, params, 0., pseudo_tmax, pseudo_dt0, mindt=pseudo_mindt, maxsteps=pseudo_maxsteps, **kwargs):
                pass
            return solver.solve(self.model, x, params, **kwargs)

    def add(self):
        self.trajectory.append(self.solution)
        for hook in self.add_hooks:
            hook.add_hook(self)

    def solve(self, params, xguess=None, **kwargs):
        params = dict(params)
        if xguess is None:
            xguess = self.solution.x
        x = self._solve(xguess, params, **kwargs)
        self.solution = steadystate_solution(
            self.model, time=self.solution.time, x=x, params=params)
        self.add()

    def _sweep(self, params, sweeps, xguess, kwargs):
        if not sweeps:
            self.solve(params, xguess=xguess, **kwargs)
            yield ((), self.state())
        else:
            sweep = sweeps[0]
            items = list(sweep(params))
            with_progress = self.progressbar(items, desc=sweep.__name__)
            for x, p in with_progress:
                first = True
                for xi, ci in self._sweep(p, sweeps[1:], xguess, kwargs):
                    if first:
                        xguess = self.x
                    first = False
                    yield (x,) + xi, ci

    def sweeps(self, params, sweeps, xguess=None, **kwargs):
        if xguess is None:
            xguess = self.solution.x
        return self._sweep(params, sweeps, self.x, kwargs)

    def sweep(self, params, sweep, **kwargs):
        for (k,), v in self.sweeps(params, [sweep], **kwargs):
            yield k, v

    def bdf1adapt(self, params, t1, dt, *args, **kwargs):
        params = dict(params)
        for time, dt, x, xt, out in solver.bdf1adapt_(
                self.model, self.solution.x, params, self.solution.time, t1, dt, *args, **kwargs):
            self.solution = bdf1_solution(
                self.model, time, x, params, (self.solution,), dt)
            self.add()
            yield self.state()

    def transientsolve(self, params, timesteps, *args, **kwargs):
        params = dict(params)
        for time, dt, x, xt, out in solver.transientsolve_(
                self.model, self.x, params, timesteps, *args, **kwargs):
            self.solution = bdf1_solution(
                self.model, time, x, params, (self.solution,), dt)
            self.add()
            yield self.state()

    def transient(self, *args, **kwargs):
        for ts in self.progressbar(self.bdf1adapt(
                *args, **kwargs), desc='transient'):
            pass

    def progressbar(self, *args, **kwargs):
        return _progressbar(*args, **kwargs)

    # AC solutions
    def acsolver(self, input_param, output_function_or_name, **kwargs):
        if callable(output_function_or_name):
            output = output_function_or_name
        else:
            assert isinstance(output_function_or_name, str)

            def output(model, time, x, xt, p, full_output):
                return full_output[output_function_or_name]
        return solver.ACSolver(self.model, self.time, self.x,
                               self.xt, self.params, input_param, output, **kwargs)

    # Misc
    def state(self):
        return context(self)


def progressbar(*args, **kwargs):
    leave = kwargs.get('leave', False)
    return progressbar_type(*args, leave=leave, **kwargs)


_progressbar = progressbar


class sweep:
    def __init__(self, parameter_name, values):
        self.parameter_name = parameter_name
        self.__name__ = parameter_name
        self._values = values

    def get_values(self, params):
        if callable(self._values):
            return self._values(params)
        else:
            return self._values

    def __call__(self, params):
        for x in self.get_values(params):
            p = dict(params)
            p[self.parameter_name] = x
            yield (x, p)


def setparam(params, pname, values):
    s = sweep(pname, values)
    return s(params)


def init_notebook():
    global progressbar_type
    progressbar_type = tqdm.tqdm_notebook



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .source import BulkSource
from oedes.ad import exp, where
import numpy as np
import scipy.constants
from oedes import functions

__all__ = ['OnsagerBraunRecombinationDissociationTerm']


class OnsagerBraunRecombinationDissociationTerm(BulkSource):
    b_eps = 1e-10
    b_max = 100

    def __init__(self, eq, electron_eq, hole_eq,
                 name='dissociation', binding_energy_param=False):
        super(
            OnsagerBraunRecombinationDissociationTerm,
            self).__init__(
                name=name,
                eq_refs=[
                    'eq',
                    'electron_eq',
                    'hole_eq'])
        assert eq.z == electron_eq.z + hole_eq.z, 'inconsistent signs'
        self.eq = eq
        self.electron_eq = electron_eq
        self.hole_eq = hole_eq
        self.binding_energy_param = binding_energy_param

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        assert eq.electron_eq.poisson is eq.hole_eq.poisson
        assert eq.electron_eq.thermal is eq.hole_eq.thermal
        poissonvars = ctx.varsOf(eq.electron_eq.poisson)
        epsilon = poissonvars['epsilon']
        Vt = ctx.varsOf(eq.electron_eq.thermal)['Vt']
        nvars = ctx.varsOf(eq.electron_eq)
        pvars = ctx.varsOf(eq.hole_eq)
        evars = ctx.varsOf(eq.eq)

        if self.binding_energy_param:
            Eb = ctx.param(eq.eq, 'binding_energy')
            a = scipy.constants.elementary_charge / \
                (Eb * 4 * np.pi * epsilon)
        else:
            a = ctx.param(eq.eq, 'distance')
            Eb = scipy.constants.elementary_charge / \
                (4 * np.pi * epsilon * a)
        u = 3. / (4. * np.pi * a**3)  # m^-3
        v = exp(-Eb / Vt)  # 1
        b = (scipy.constants.elementary_charge / (8 * np.pi)) * \
            poissonvars['Ecellm'] / (epsilon * Vt**2)  # 1
        # b=0.
        t = functions.OnsagerFunction(
            where(
                b < self.b_max,
                b,
                self.b_max) +
            self.b_eps)  # 1
        gamma = scipy.constants.elementary_charge * \
            (nvars['mu_cell'] + pvars['mu_cell']) / \
            epsilon  # m^3 /s
        r = gamma * (nvars['c'] * pvars['c'] - ctx.common_param(
            [eq.electron_eq, eq.hole_eq], 'npi'))  # 1/(m^3 s) # TODO
        d = gamma * evars['c'] * u * v * t
        ctx.outputCell([eq.eq, self.name, 'recombination'],
                       r, unit=ctx.units.dconcentration_dt)
        ctx.output([eq.eq, self.name, 'dissociation'],
                   d, unit=ctx.units.dconcentration_dt)
        f = r - d
        self.add(ctx, f, plus=[eq.eq], minus=[eq.hole_eq, eq.electron_eq])
